import unittest
import uuid

from vdv736.isotime import timestamp
from vdv736.model import Subscription
from vdv736.request import SiriRequest
from vdv736.request import SubscriptionRequest
from vdv736.request import SituationExchangeSubscriptionRequest

from vdv736.sirixml import get_value as sirixml_get_value
from vdv736.sirixml import get_attribute as sirixml_get_attribue


class SiriRequest_Test(unittest.TestCase):
    def test_SiriRequest(self):

        request = SiriRequest()

        self.assertIsNotNone(request.xml())
        self.assertEqual(sirixml_get_attribue(request, 'Siri.version'), '2.0')


class SubscriptionRequest_Test(unittest.TestCase):
    def test_SubscriptionRequest(self):

        subscriber_ref = 'PY_TEST_SUBSCRIBER'
        
        request = SubscriptionRequest(
            subscriber_ref
        )
        
        self.assertIsNotNone(request.xml())
        self.assertEqual(sirixml_get_value(request, 'Siri.SubscriptionRequest.RequestorRef'), subscriber_ref)


class SituationExchangeSubscriptionRequest_Test(unittest.TestCase):
    def test_SituationExchangeSubscriptionRequest(self):

        subscriber_ref = 'PY_TEST_SUBSCRIBER'
        subscription_id = str(uuid.uuid4())
        subscription_termination = timestamp(60 * 60 * 24)
        
        subscription = Subscription.create(
            subscription_id, 
            'http://127.0.0.1', 
            8080, 
            'https',
            subscriber_ref, 
            timestamp(60 * 60 * 24)
        )

        request = SituationExchangeSubscriptionRequest(subscription)
        
        self.assertIsNotNone(request.xml())
        self.assertEqual(sirixml_get_value(request, 'Siri.SubscriptionRequest.SituationExchangeSubscriptionRequest.SubscriberRef'), subscriber_ref)
        self.assertEqual(sirixml_get_value(request, 'Siri.SubscriptionRequest.SituationExchangeSubscriptionRequest.SubscriptionIdentifier'), subscription_id)
        self.assertEqual(sirixml_get_value(request, 'Siri.SubscriptionRequest.SituationExchangeSubscriptionRequest.InitialTerminationTime'), subscription_termination)




import logging
import requests
import uuid
import time
import typing
import uvicorn

from .isotime import timestamp
from .database import local_node_database
from .datalog import Datalog
from .delivery import xml2siri_delivery
from .delivery import SiriDelivery
from .delivery import SituationExchangeDelivery
from .handler import SituationProgressHandler
from .model import PublicTransportSituation
from .model import Subscription
from .participantconfig import ParticipantConfig
from .request import InvalidMethodError
from .request import SiriRequest
from .request import CheckStatusRequest
from .request import SituationExchangeSubscriptionRequest
from .request import TerminateSubscriptionRequest
from .request import SituationExchangeRequest
from .response import xml2siri_response
from .response import SiriResponse
from .response import DataReceivedAcknowledgement
from .sirixml import exists as sirixml_exists
from .sirixml import get_elements as sirixml_get_elements
from .sirixml import get_value as sirixml_get_value

from fastapi import FastAPI
from fastapi import BackgroundTasks
from fastapi import APIRouter
from fastapi import Request
from fastapi import Response
from threading import Thread


class Subscriber():

    def __init__(self, participant_ref: str, participant_config_filename: str, local_ip_address: str = '0.0.0.0', publish_subscribe: bool = True, datalog_directory: str|None = None):
        self._service_participant_ref = participant_ref
        self._service_local_ip_address = local_ip_address
        self._pubsub = publish_subscribe
        self._datalog = datalog_directory

        self._logger = logging.getLogger('uvicorn')

        self._local_node_database = local_node_database('vdv736.subscriber')
        self._endpoint = None

        self._on_delivery = None

        try:
            self._participant_config = ParticipantConfig(participant_config_filename)
        except Exception as ex:
            self._logger.error(ex)

    def __enter__(self):

        if self._pubsub:
            self._endpoint_thread = Thread(target=self._run_endpoint, args=(), daemon=True)
            self._endpoint_thread.start()

            time.sleep(0.01) # give the endpoint thread time for startup
            self._logger.info(f"Subscriber running at {self._participant_config.participants[self._service_participant_ref]['host']}:{self._participant_config.participants[self._service_participant_ref]['port']}")
            self._logger.info(f"Local node database at {self._local_node_database._filename}")

        return self

    def __exit__(self, exception_type, exception_value, exception_traceback) -> None:
        
        if self._pubsub:
            # terminate all subscriptions
            for subscription_id, subscription in self._local_node_database.get_subscriptions().items():
                self.unsubscribe(subscription_id)

            # terminate endpoint and close local database
            if self._endpoint is not None:
                self._endpoint.terminate()
            
            if self._endpoint_thread is not None:
                self._endpoint_thread.join(1)

        # local node database has to be closed anyway, regardless of using pubsub mode
        if self._local_node_database is not None:
            self._local_node_database.close(True)

    def set_callbacks(self, on_delivery_callback: typing.Callable[[SiriDelivery], None]) -> None:
        self._on_delivery = on_delivery_callback

        if self._endpoint is not None:
            self._endpoint.set_callbacks(self._on_delivery)

    def get_situations(self) -> dict[str, PublicTransportSituation]:
        return self._local_node_database.get_situations()

    def status(self, subscription_id=None) -> bool:
        if not self._pubsub:
            raise RuntimeError("Status requests are only available in publish/subscribe mode!")
        
        if subscription_id is not None:
            return self._status(subscription_id)
        else:
            all_subscriptions_ok = True
            for subscription_id, _ in self._local_node_database.get_subscriptions().items():
                if self._status(subscription_id) != True:
                    all_subscriptions_ok = False

            return all_subscriptions_ok
        
    def _status(self, subscription_id: str) -> bool:
        subscription = self._local_node_database.get_subscriptions()[subscription_id]

        request = CheckStatusRequest(subscription)
        response = self._send_request(subscription, request)

        if response is not None and sirixml_get_value(response, 'Siri.CheckStatusResponse.Status', False):
            if subscription.remote_service_startup_time is not None:
                if sirixml_get_value(response, 'Siri.CheckStatusResponse.ServiceStartedTime') == subscription.remote_service_startup_time:
                    self._logger.info(f"Status for subscription {subscription.id} @ {subscription.remote_service_participant_ref} as {subscription.subscriber} OK")
                    return True
                else:
                    self._logger.warning(f"Remote server for subscription {subscription.id} @ {subscription.remote_service_participant_ref} as {subscription.subscriber} seems to be restarted")
                    
                    self.unsubscribe(subscription.id)
                    return self.subscribe(subscription.remote_service_participant_ref) is not None
            else:
                subscription.remote_service_startup_time = sirixml_get_value(response, 'Siri.CheckStatusResponse.ServiceStartedTime')
                self._local_node_database.update_subscription(subscription_id, subscription)

                self._logger.info(f"Status for subscription {subscription.id} @ {subscription.remote_service_participant_ref} as {subscription.subscriber} OK")
                return True
        else:
            self._logger.error(f"Status for subscription {subscription.id} @ {subscription.remote_service_participant_ref} as {subscription.subscriber} FAIL")
            return False

    def subscribe(self, participant_ref: str) -> str|None:

        if not self._pubsub:
            raise RuntimeError("Subscriptions can only be initialized in publish/subscribe mode!")
        
        subscription_id = str(uuid.uuid4())
        subscription_host = self._participant_config.participants[participant_ref]['host']
        subscription_port = self._participant_config.participants[participant_ref]['port']
        subscription_protocol = self._participant_config.participants[participant_ref]['protocol']
        subscription_termination = timestamp(60 * 60 * 24)

        subscription = Subscription.create(subscription_id, subscription_host, subscription_port, subscription_protocol, self._service_participant_ref, subscription_termination)
        subscription.single_endpoint = self._participant_config.participants[participant_ref]['single_endpoint']
        subscription.status_endpoint = self._participant_config.participants[participant_ref]['status_endpoint']
        subscription.subscribe_endpoint = self._participant_config.participants[participant_ref]['subscribe_endpoint']
        subscription.unsubscribe_endpoint = self._participant_config.participants[participant_ref]['unsubscribe_endpoint']

        subscription.remote_service_participant_ref = participant_ref

        request = SituationExchangeSubscriptionRequest(subscription)
        response = self._send_request(subscription, request)

        if response is not None and sirixml_get_value(response, 'Siri.SubscriptionResponse.ResponseStatus.Status', True):
            self._logger.info(f"Initialized subscription {subscription.id} @ {subscription.remote_service_participant_ref} as {subscription.subscriber} successfully")

            service_started_time = sirixml_get_value(response, 'Siri.SubscriptionResponse.ResponseStatus.ServiceStartedTime')
            if service_started_time is not None:
                subscription.remote_service_startup_time = service_started_time
                
            self._local_node_database.add_subscription(subscription_id, subscription)

            return subscription_id
        else:
            self._logger.error(f"Failed to initalize subscription {subscription.id} @ {subscription.remote_service_participant_ref} as {subscription.subscriber}")

            return None
        
    def unsubscribe(self, subscription_id: str) -> bool:
        
        if not self._pubsub:
            raise RuntimeError("Subscriptions can only be terminated in publish/subscribe mode!")
        
        # take subscription instance from subscription stack
        subscription = self._local_node_database.get_subscriptions()[subscription_id]

        # delete subscription out of local database
        self._local_node_database.remove_subscription(subscription_id)
        
        # create termination request here ...
        request = TerminateSubscriptionRequest(self._service_participant_ref)
        request.subscription(self._service_participant_ref, subscription_id)

        response = self._send_request(subscription, request)

        # check each termination subscription response for success
        if response is not None:
            if sirixml_exists(response, 'Siri.TerminationSubscriptionResponse.TerminationResponseStatus'):
                for termination_response_status in sirixml_get_elements(response, 'Siri.TerminationSubscriptionResponse.TerminationResponseStatus'):
                    if termination_response_status.Status == True:
                        self._logger.info(f"Terminated subscription {subscription.id} @ {subscription.remote_service_participant_ref} as {subscription.subscriber} successfully")
                        return True
                    else:
                        self._logger.error(f"Failed to terminate subscription {subscription.id} @ {subscription.remote_service_participant_ref} as {subscription.subscriber}")
                        return False
            else:
                # publisher returns no termination status at all, that means, there were no subscriptions at publisher side ... good anyway
                self._logger.info(f"Terminated subscription {subscription.id} @ {subscription.remote_service_participant_ref} as {subscription.subscriber} successfully")
                return True
        else:
            self._logger.error(f"Failed to terminate subscription {subscription.id} @ {subscription.remote_service_participant_ref} as {subscription.subscriber}")
            
    def request(self, publisher_ref: str, method: str = 'POST', headers: dict = dict()) -> bool:

        if self._pubsub:
            raise RuntimeError("Direct requests are only available in request/response mode!")

        # generate SituationExchangeRequest
        request = SituationExchangeRequest(self._service_participant_ref)
        delivery = self._send_direct_request(publisher_ref, request, method, headers)

        if delivery is not None:
            # check whether on_delivery callback is used ...
            if self._on_delivery is not None:
                self._on_delivery(delivery)

            # process service delivery ...
            for pts in sirixml_get_elements(delivery, 'Siri.ServiceDelivery.SituationExchangeDelivery.Situations.PtSituationElement'):
                situation_id = sirixml_get_value(pts, 'SituationNumber')
                
                result = SituationProgressHandler.handle_situation(pts)

                if result:
                    self._local_node_database.add_or_update_situation(situation_id, pts)
                else:
                    self._local_node_database.remove_situation(situation_id)

            return True
        else:
            self._logger.error(f"Failed to request data from {publisher_ref}")

            return False

    def _run_endpoint(self) -> None:
        self._endpoint = SubscriberEndpoint(self._service_participant_ref, self._datalog)

        # disable uvicorn logs
        logging.getLogger('uvicorn.error').handlers = []
        logging.getLogger('uvicorn.error').propagate = False

        logging.getLogger('uvicorn.access').handlers = []
        logging.getLogger('uvicorn.access').propagate = False

        logging.getLogger('uvicorn.asgi').handlers = []
        logging.getLogger('uvicorn.asgi').propagate = False

        # run ASGI server with endpoint
        endpoint_host = self._service_local_ip_address # self._participant_config.participants[self._service_participant_ref]['host']
        endpoint_port = self._participant_config.participants[self._service_participant_ref]['port']

        uvicorn.run(app=self._endpoint.create_endpoint(
            self._participant_config.participants[self._service_participant_ref]['single_endpoint'],
            self._participant_config.participants[self._service_participant_ref]['delivery_endpoint']
        ), host=endpoint_host, port=endpoint_port)

    def _send_request(self, subscription: Subscription, siri_request: SiriRequest) -> SiriResponse|None:
        try:
            if isinstance(siri_request, CheckStatusRequest):
                status_endpoint = subscription.single_endpoint if subscription.single_endpoint is not None else subscription.status_endpoint
                endpoint = f"{subscription.protocol}://{subscription.host}:{subscription.port}{status_endpoint}"
            elif isinstance(siri_request, SituationExchangeSubscriptionRequest):
                subscribe_endpoint = subscription.single_endpoint if subscription.single_endpoint is not None else subscription.subscribe_endpoint
                endpoint = f"{subscription.protocol}://{subscription.host}:{subscription.port}{subscribe_endpoint}"
            elif isinstance(siri_request, TerminateSubscriptionRequest):
                unsubscribe_endpoint = subscription.single_endpoint if subscription.single_endpoint is not None else subscription.unsubscribe_endpoint
                endpoint = f"{subscription.protocol}://{subscription.host}:{subscription.port}{unsubscribe_endpoint}"
            
            headers = {
                "Content-Type": "application/xml"
            }
            
            if self._datalog is not None:
                Datalog.create(self._datalog, siri_request.xml(), {
                    'method': 'POST',
                    'endpoint': endpoint,
                    'headers': headers
                }, self._service_participant_ref, 'OUT', type(siri_request).__name__, 'Request')

            response_xml = requests.post(endpoint, headers=headers, data=siri_request.xml())

            if self._datalog is not None:
                Datalog.create(self._datalog, response_xml.content, {
                    'method': 'POST',
                    'endpoint': endpoint,
                    'headers': headers
                }, self._service_participant_ref, 'OUT', type(siri_request).__name__, 'Response')

            response = xml2siri_response(response_xml.content)

            return response
        except Exception as ex:
            self._logger.error(ex)
            return None
        
    def _send_direct_request(self, publisher_ref: str, siri_request: SiriRequest, method: str, headers: dict = dict()) -> SituationExchangeDelivery|None:
        try:
            subscription_host = self._participant_config.participants[publisher_ref]['host']
            subscription_port = self._participant_config.participants[publisher_ref]['port']
            subscription_protocol = self._participant_config.participants[publisher_ref]['protocol']
            
            if isinstance(siri_request, SituationExchangeRequest):
                request_endpoint = self._participant_config.participants[publisher_ref]['single_endpoint'] if self._participant_config.participants[publisher_ref]['single_endpoint'] is not None else self._participant_config.participants[publisher_ref]['request_endpoint']
                endpoint = f"{subscription_protocol}://{subscription_host}:{subscription_port}{request_endpoint}"
            
            siri_headers = {
                "Content-Type": "application/xml"
            }

            if len(headers) > 0:
                siri_headers = siri_headers | headers
            
            if self._datalog is not None:
                Datalog.create(self._datalog, siri_request.xml(), {
                    'method': method.upper(),
                    'endpoint': endpoint,
                    'headers': headers
                }, self._service_participant_ref, 'OUT', type(siri_request).__name__, 'Request')

            if method.lower() == 'post':
                response_xml = requests.post(endpoint, headers=siri_headers, data=siri_request.xml())
            elif method.lower() == 'get':
                response_xml = requests.get(endpoint, headers=siri_headers)
            else:
                raise InvalidMethodError(f"Invalid request method {method}!")

            if self._datalog is not None:
                Datalog.create(self._datalog, response_xml.content, {
                    'method': method.upper(),
                    'endpoint': endpoint,
                    'headers': headers
                }, self._service_participant_ref, 'OUT', type(siri_request).__name__, 'Response')

            delivery = xml2siri_delivery(response_xml.content)

            return delivery
        except Exception as ex:

            # re-throw occuring InvalidMethodError in this case
            if isinstance(ex, InvalidMethodError):
                raise ex
            else:
                self._logger.error(ex)

            return None


class SubscriberEndpoint():

    def __init__(self, participant_ref: str, datalog_directory: str|None = None):
        self._service_participant_ref = participant_ref
        self._service_startup_time = timestamp()
        self._logger = logging.getLogger('uvicorn')
        self._datalog = datalog_directory

        self._router = APIRouter()
        self._endpoint = FastAPI()

        self._local_node_database = local_node_database('vdv736.subscriber')

        self._on_delivery = None

    def set_callbacks(self, on_delivery_callback: typing.Callable[[SiriDelivery], None]|None) -> None:
        self._on_delivery = on_delivery_callback
    
    def create_endpoint(self, single_endpoint: str|None = None, delivery_endpoint: str = '/delivery') -> FastAPI:
        if single_endpoint is not None:
            self._router.add_api_route(single_endpoint, self._dispatcher, methods=['POST'])
        else:
            self._router.add_api_route(delivery_endpoint, self._delivery, methods=['POST'])
        
        self._endpoint.include_router(self._router)

        return self._endpoint
    
    def terminate(self) -> None:
        self._local_node_database.close()
    
    async def _dispatcher(self, req: Request, bgt: BackgroundTasks) -> Response:
        body = str(await req.body())

        if '<ServiceDelivery' in body:
            return await self._delivery(req, bgt)
        else:
            return Response(status_code=400)

    async def _delivery(self, req: Request, bgt: BackgroundTasks) -> Response:
        try:
            xml = await req.body()
            
            if self._datalog is not None:
                Datalog.create(self._datalog, xml, {
                    'method': req.method,
                    'endpoint': str(req.url),
                    'headers': dict(req.headers)
                }, self._service_participant_ref, 'IN', 'SituationExchangeDelivery', 'Request')
            
            delivery = xml2siri_delivery(xml)

            # check for active subscriptions from the publisher who has sent the delivery
            delivery_producer_ref = sirixml_get_value(delivery, 'Siri.ServiceDelivery.ProducerRef')
            
            subscription_at_producer_found = False
            for _, subscription in self._local_node_database.get_subscriptions().items():
                if subscription.remote_service_participant_ref == delivery_producer_ref:
                    subscription_at_producer_found = True
                    break

            if not subscription_at_producer_found:
                return Response(status_code=401)
            
            # process service delivery ...
            for pts in sirixml_get_elements(delivery, 'Siri.ServiceDelivery.SituationExchangeDelivery.Situations.PtSituationElement'):
                situation_id = sirixml_get_value(pts, 'SituationNumber')
                
                result = SituationProgressHandler.handle_situation(pts)

                if result:
                    self._local_node_database.add_or_update_situation(situation_id, pts)
                else:
                    self._local_node_database.remove_situation(situation_id)

            # create data acknowledgement with OK status
            acknowledgement = DataReceivedAcknowledgement(
                sirixml_get_value(delivery, 'Siri.ServiceDelivery.SituationExchangeDelivery.SubscriberRef'), 
                sirixml_get_value(delivery, 'Siri.ServiceDelivery.ResponseMessageIdentifier')
            )

            acknowledgement.ok()

            # run callback method for delivery
            if self._on_delivery is not None:
                bgt.add_task(self._on_delivery, delivery)

            if self._datalog is not None:
                Datalog.create(self._datalog, acknowledgement.xml(), {
                    'method': req.method,
                    'endpoint': str(req.url),
                    'headers': dict(req.headers)
                }, self._service_participant_ref, 'IN', 'SituationExchangeDelivery', 'Response')

            return Response(content=acknowledgement.xml(), media_type='application/xml')
        except Exception as ex:
            self._logger.error(ex)

            # create data acknowledgement with Fail status
            acknowledgement = DataReceivedAcknowledgement(
                sirixml_get_value(delivery, 'Siri.ServiceDelivery.SituationExchangeDelivery.SubscriberRef'), 
                sirixml_get_value(delivery, 'Siri.ServiceDelivery.ResponseMessageIdentifier')
            )

            acknowledgement.error()

            return Response(content=acknowledgement.xml(), media_type='application/xml')
        



import logging
import uuid
import sys
import time

from vdv736.model import PublicTransportSituation
from vdv736.publisher import Publisher

logging.basicConfig(level=logging.INFO, format="[%(asctime)s] %(levelname)s: %(message)s")

with Publisher('PY_TEST_PUBLISHER', './demo/demo_participants.yaml') as publisher:
    
    time.sleep(5)
    publisher.publish_situation(PublicTransportSituation.create(uuid.uuid4()))
    time.sleep(15)
    publisher.publish_situation(PublicTransportSituation.create(uuid.uuid4()))
    time.sleep(40)
    publisher.publish_situation(PublicTransportSituation.create(uuid.uuid4()))

    while True:
        publisher.publish_situation(PublicTransportSituation.create(uuid.uuid4()))
        time.sleep(60)  



import os
import unittest

from vdv736.participantconfig import ParticipantConfig

class SiriRequest_Test(unittest.TestCase):

    def test_withFullConfig(self):

        participant_config = ParticipantConfig(
            os.path.join(os.path.dirname(__file__), 'data/yaml/participants_with_full_config.yaml')
        )

        self.assertEqual(2, len(participant_config.participants))
        self.assertEqual('127.0.0.1', participant_config.participants['PY_TEST_PUBLISHER']['host'])
        self.assertEqual(9091, participant_config.participants['PY_TEST_PUBLISHER']['port'])
        self.assertEqual('http', participant_config.participants['PY_TEST_PUBLISHER']['protocol'])
        self.assertEqual(None, participant_config.participants['PY_TEST_PUBLISHER']['single_endpoint'])
        self.assertEqual('/status', participant_config.participants['PY_TEST_PUBLISHER']['status_endpoint'])
        self.assertEqual('/subscribe', participant_config.participants['PY_TEST_PUBLISHER']['subscribe_endpoint'])
        self.assertEqual('/unsubscribe', participant_config.participants['PY_TEST_PUBLISHER']['unsubscribe_endpoint'])
        self.assertEqual('/request', participant_config.participants['PY_TEST_PUBLISHER']['request_endpoint'])
        self.assertEqual(None, participant_config.participants['PY_TEST_PUBLISHER']['delivery_endpoint'])

    def test_withPartialConfig(self):

        participant_config = ParticipantConfig(
            os.path.join(os.path.dirname(__file__), 'data/yaml/participants_with_partial_config.yaml')
        )

        self.assertEqual(2, len(participant_config.participants))
        self.assertEqual('127.0.0.1', participant_config.participants['PY_TEST_SUBSCRIBER']['host'])
        self.assertEqual(9090, participant_config.participants['PY_TEST_SUBSCRIBER']['port'])
        self.assertEqual('http', participant_config.participants['PY_TEST_SUBSCRIBER']['protocol'])
        self.assertEqual('/server', participant_config.participants['PY_TEST_SUBSCRIBER']['single_endpoint'])
        self.assertEqual('/status', participant_config.participants['PY_TEST_SUBSCRIBER']['status_endpoint'])
        self.assertEqual('/subscribe', participant_config.participants['PY_TEST_SUBSCRIBER']['subscribe_endpoint'])
        self.assertEqual('/unsubscribe', participant_config.participants['PY_TEST_SUBSCRIBER']['unsubscribe_endpoint'])
        self.assertEqual('/request', participant_config.participants['PY_TEST_SUBSCRIBER']['request_endpoint'])
        self.assertEqual('/delivery', participant_config.participants['PY_TEST_SUBSCRIBER']['delivery_endpoint'])

    def test_withInvalidConfig(self):
        self.assertRaises(ValueError, ParticipantConfig, os.path.join(os.path.dirname(__file__), 'data/yaml/participants_with_invalid_config.yaml'))


import logging
import requests
import time
import typing
import uvicorn

from .isotime import timestamp
from .database import local_node_database
from .datalog import Datalog
from .delivery import ServiceDelivery
from .delivery import SituationExchangeDelivery
from .model import PublicTransportSituation
from .model import Subscription
from .participantconfig import ParticipantConfig
from .request import xml2siri_request
from .response import xml2siri_response
from .response import SiriResponse
from .response import CheckStatusResponse
from .response import SubscriptionResponse
from .response import TerminateSubscriptionResponse
from .sirixml import get_elements as sirixml_get_elements
from .sirixml import get_value as sirixml_get_value

from fastapi import FastAPI
from fastapi import BackgroundTasks
from fastapi import APIRouter
from fastapi import Request
from fastapi import Response
from threading import Thread


class Publisher():

    def __init__(self, participant_ref: str, participant_config_filename: str, local_ip_address: str = '0.0.0.0', datalog_directory: str|None = None):
        self._service_participant_ref = participant_ref
        self._service_local_ip_address = local_ip_address
        self._datalog = datalog_directory

        self._logger = logging.getLogger('uvicorn')

        self._local_node_database = local_node_database('vdv736.publisher')

        try:
            self._participant_config = ParticipantConfig(participant_config_filename)
        except Exception as ex:
            self._logger.error(ex)

        self._on_subscribe = None
        self._on_unsubscribe = None

    def __enter__(self):
        
        self._endpoint_thread = Thread(target=self._run_endpoint, args=(), daemon=True)
        self._endpoint_thread.start()

        time.sleep(0.01) # give the endpoint thread time for startup
        self._logger.info(f"Publisher running at {self._participant_config.participants[self._service_participant_ref]['host']}:{self._participant_config.participants[self._service_participant_ref]['port']}")
        self._logger.info(f"Local node database at {self._local_node_database._filename}")

        # set internal callbacks
        self._endpoint.set_callbacks(
            on_subscribe_callback=self._on_subscribe_internal,
            on_unsubscribe_callback=self._on_unsubscribe_internal
        )

        return self

    def __exit__(self, exception_type, exception_value, exception_traceback) -> None:
        if self._endpoint is not None:
            self._endpoint.terminate()
        
        if self._endpoint_thread is not None:
            self._endpoint_thread.join(1)

        if self._local_node_database is not None:
            self._local_node_database.close(True)
    
    def set_callbacks(self, on_status_callback: typing.Callable[[], None]|None = None, on_subscribe_callback: typing.Callable[[Subscription], None]|None = None, on_unsubscribe_callback: typing.Callable[[Subscription], None]|None = None, on_request_callback: typing.Callable[[], None]|None = None) -> None:
        self._on_subscribe = on_subscribe_callback
        self._on_unsubscribe = on_unsubscribe_callback

        self._endpoint.set_callbacks(
            on_status_callback=on_status_callback,
            on_subscribe_callback=self._on_subscribe_internal,
            on_unsubscribe_callback=self._on_unsubscribe_internal, 
            on_request_callback=on_request_callback
        )

    def publish_situation(self, situation: PublicTransportSituation) -> None:
        situation_id = sirixml_get_value(situation, 'SituationNumber')
        self._local_node_database.add_or_update_situation(situation_id, situation)
        
        for _, subscription in self._local_node_database.get_subscriptions().items():
            delivery = SituationExchangeDelivery(self._service_participant_ref, subscription)
            delivery.add_situation(situation)

            response = self._send_request(subscription, delivery)

            if response is not None and sirixml_get_value(response, 'Siri.DataReceivedAcknowledgement.Status', False):
                self._logger.info(f"Sent delivery for subscription {subscription.id} to {subscription.subscriber} successfully")
            else:
                self._logger.error(f"Failed to send delivery for subscription {subscription.id} to {subscription.subscriber}")

    def _on_subscribe_internal(self, subscription: Subscription) -> None:
        
        # send initial load here
        delivery = SituationExchangeDelivery(self._service_participant_ref, subscription)
        for _, situation in self._local_node_database.get_situations().items():
            delivery.add_situation(situation)

        response = self._send_request(subscription, delivery)

        if sirixml_get_value(response, 'Siri.DataReceivedAcknowledgement.Status', False):
            self._logger.info(f"Sent initial load delivery for subscription {subscription.id} to {subscription.subscriber} successfully")
        else:
            self._logger.error(f"Failed to send initial load delivery for subscription {subscription.id} to {subscription.subscriber}")

        # call external callback method
        if self._on_subscribe is not None:
            self._on_subscribe(subscription)

    def _on_unsubscribe_internal(self, subscription: Subscription) -> None:

        # call external callback method
        if self._on_unsubscribe is not None:
            self._on_unsubscribe(subscription)
    
    def _run_endpoint(self) -> None:
        self._endpoint = PublisherEndpoint(self._service_participant_ref, self._datalog)

        # disable uvicorn logs
        logging.getLogger('uvicorn.error').handlers = []
        logging.getLogger('uvicorn.error').propagate = False

        logging.getLogger('uvicorn.access').handlers = []
        logging.getLogger('uvicorn.access').propagate = False

        logging.getLogger('uvicorn.asgi').handlers = []
        logging.getLogger('uvicorn.asgi').propagate = False

        # run ASGI server with endpoint
        endpoint_host = self._service_local_ip_address # self._participant_config.participants[self._service_participant_ref]['host']
        endpoint_port = self._participant_config.participants[self._service_participant_ref]['port']

        uvicorn.run(app=self._endpoint.create_endpoint(
            self._service_participant_ref,
            self._participant_config.participants[self._service_participant_ref]['single_endpoint'],
            self._participant_config.participants[self._service_participant_ref]['status_endpoint'],
            self._participant_config.participants[self._service_participant_ref]['subscribe_endpoint'],
            self._participant_config.participants[self._service_participant_ref]['unsubscribe_endpoint'],
            self._participant_config.participants[self._service_participant_ref]['request_endpoint']
        ), host=endpoint_host, port=endpoint_port)

    def _send_request(self, subscription: Subscription, siri_delivery: ServiceDelivery) -> SiriResponse|None:
        try:
            subscription_host = self._participant_config.participants[subscription.subscriber]['host']
            subscription_port = self._participant_config.participants[subscription.subscriber]['port']
            subscription_protocol = self._participant_config.participants[subscription.subscriber]['protocol']
            
            if isinstance(siri_delivery, SituationExchangeDelivery):
                delivery_endpoint = self._participant_config.participants[subscription.subscriber]['single_endpoint'] if self._participant_config.participants[subscription.subscriber]['single_endpoint'] is not None else self._participant_config.participants[subscription.subscriber]['delivery_endpoint']
                endpoint = f"{subscription_protocol}://{subscription_host}:{subscription_port}{delivery_endpoint}"

            headers = {
                "Content-Type": "application/xml"
            }

            if self._datalog is not None:
                Datalog.create(self._datalog, siri_delivery.xml(), {
                    'method': 'POST',
                    'endpoint': endpoint,
                    'headers': headers
                }, self._service_participant_ref, 'OUT', type(siri_delivery).__name__, 'Request')
            
            response_xml = requests.post(endpoint, headers=headers, data=siri_delivery.xml())

            if self._datalog is not None:
                Datalog.create(self._datalog, response_xml.content, {
                    'method': 'POST',
                    'endpoint': endpoint,
                    'headers': headers
                }, self._service_participant_ref, 'OUT', type(siri_delivery).__name__, 'Response')

            response = xml2siri_response(response_xml.content)

            return response
        except Exception as ex:
            self._logger.exception(ex)
            return None


class PublisherEndpoint():

    def __init__(self, participant_ref: str, datalog_directory: str|None = None):
        self._service_participant_ref = participant_ref
        self._service_startup_time = timestamp()
        self._logger = logging.getLogger('uvicorn')
        self._datalog = datalog_directory

        self._router = APIRouter()
        self._endpoint = FastAPI()

        self._local_node_database = local_node_database('vdv736.publisher')
        
        self._on_status = None
        self._on_subscribe = None
        self._on_unsubscribe = None
        self._on_request = None

    def set_callbacks(self, on_status_callback: typing.Callable[[], None]|None = None, on_subscribe_callback: typing.Callable[[Subscription], None]|None = None, on_unsubscribe_callback: typing.Callable[[Subscription], None]|None = None, on_request_callback: typing.Callable[[], None]|None = None) -> None:
        self._on_status = on_status_callback
        self._on_subscribe = on_subscribe_callback
        self._on_unsubscribe = on_unsubscribe_callback
        self._on_request = on_request_callback

    def create_endpoint(self, participant_ref: str, single_endpoint: str|None = None, status_endpoint: str = '/status', subscribe_endpoint: str = '/subscribe', unsubscribe_endpoint: str = '/unsubscribe', request_endpoint: str = '/request') -> FastAPI:
        self._participant_ref = participant_ref

        if single_endpoint is not None:
            self._router.add_api_route(single_endpoint, self._dispatcher, methods=['POST'])
        else:
            self._router.add_api_route(status_endpoint, self._status, methods=['POST'])
            self._router.add_api_route(subscribe_endpoint, self._subscribe, methods=['POST'])
            self._router.add_api_route(unsubscribe_endpoint, self._unsubscribe, methods=['POST'])
            self._router.add_api_route(request_endpoint, self._request, methods=['POST'])
        
        self._endpoint.include_router(self._router)

        return self._endpoint
    
    def terminate(self):
        self._local_node_database.close()
    
    async def _dispatcher(self, req: Request, bgt: BackgroundTasks) -> Response:
        body = str(await req.body())

        if '<CheckStatusRequest' in body:
            return await self._status(req, bgt)
        elif '<SubscriptionRequest' in body:
            return await self._subscribe(req, bgt)
        elif '<TerminateSubscriptionRequest' in body:
            return await self._unsubscribe(req, bgt)
        elif '<SituationExchangeRequest' in body:
            return await self._request(req, bgt)
        else:
            return Response(status_code=400)

    async def _status(self, req: Request, bgt: BackgroundTasks) -> Response:
        xml = await req.body()
            
        if self._datalog is not None:
            Datalog.create(self._datalog, xml, {
                'method': req.method,
                'endpoint': str(req.url),
                'headers': dict(req.headers)
            }, self._service_participant_ref, 'IN', 'CheckStatusRequest', 'Request')
        
        request = xml2siri_request(xml)

        # run callback method for status
        if self._on_status is not None:
            bgt.add_task(self._on_status)

        # simply respond with current status
        response = CheckStatusResponse(self._service_startup_time)

        if self._datalog is not None:
            Datalog.create(self._datalog, response.xml(), {
                'method': req.method,
                'endpoint': str(req.url),
                'headers': dict(req.headers)
            }, self._service_participant_ref, 'IN', 'CheckStatusRequest', 'Response')

        return Response(content=response.xml(), media_type='application/xml')

    async def _subscribe(self, req: Request, bgt: BackgroundTasks) -> Response:
        xml = await req.body()
            
        if self._datalog is not None:
            Datalog.create(self._datalog, xml, {
                'method': req.method,
                'endpoint': str(req.url),
                'headers': dict(req.headers)
            }, self._service_participant_ref, 'IN', 'SituationExchangeSubscriptionRequest', 'Request')
        
        request = xml2siri_request(xml)

        # add subscription parameters to subscription index
        subscription_id = sirixml_get_value(request, 'Siri.SubscriptionRequest.SituationExchangeSubscriptionRequest.SubscriptionIdentifier')
        subscription_termination = sirixml_get_value(request, 'Siri.SubscriptionRequest.SituationExchangeSubscriptionRequest.InitialTerminationTime')

        subscription = Subscription.create(
            subscription_id,
            None,
            None,
            None,
            sirixml_get_value(request, 'Siri.SubscriptionRequest.RequestorRef'),
            subscription_termination
        )
            
        try:
            result = self._local_node_database.add_subscription(subscription_id, subscription)

            # respond with SubscriptionResponse OK
            response = SubscriptionResponse(self._participant_ref, self._service_startup_time)

            if result == True:
                response.ok(subscription_id, subscription_termination)

                # run callback method for subscriptions
                if self._on_subscribe is not None:
                    bgt.add_task(self._on_subscribe, subscription)
            else:
                response.error(subscription_id)

            if self._datalog is not None:
                Datalog.create(self._datalog, response.xml(), {
                    'method': req.method,
                    'endpoint': str(req.url),
                    'headers': dict(req.headers)
                }, self._service_participant_ref, 'IN', 'SituationExchangeSubscriptionRequest', 'Response')

            return Response(content=response.xml(), media_type='application/xml')
        except Exception as ex:
            # log exception
            self._logger.error(ex)

            # respond with SubscriptionResponse Error
            response = SubscriptionResponse(self._participant_ref, self._service_startup_time)
            response.error(subscription_id)

            if self._datalog is not None:
                Datalog.create(self._datalog, response.xml(), {
                    'method': req.method,
                    'endpoint': str(req.url),
                    'headers': dict(req.headers)
                }, self._service_participant_ref, 'IN', 'SituationExchangeSubscriptionRequest', 'Response')

            return Response(content=response.xml(), media_type='application/xml')

    async def _unsubscribe(self, req: Request, bgt: BackgroundTasks) -> Response:
        xml = await req.body()
            
        if self._datalog is not None:
            Datalog.create(self._datalog, xml, {
                'method': req.method,
                'endpoint': str(req.url),
                'headers': dict(req.headers)
            }, self._service_participant_ref, 'IN', 'TerminateSubscriptionRequest', 'Request')
        
        request = xml2siri_request(xml)

        subscriber_ref = sirixml_get_value(request, 'Siri.TerminateSubscriptionRequest.RequestorRef')

        # check which subscription should be deleted - currently, only all subscriptions by a certain subscriber can be deleted
        subscriptions_to_delete = list()

        for subscription_id, subscription in self._local_node_database.get_subscriptions().items():
            if subscription.subscriber == subscriber_ref:
                subscriptions_to_delete.append(subscription_id)

        response = TerminateSubscriptionResponse(self._participant_ref)
        for subscription_id in subscriptions_to_delete:
            try:
                # delete subscription from subscription stack
                result = self._local_node_database.remove_subscription(subscription_id)

                # respond with SubscriptionResponse OK or ERROR depending on result
                if result == True:
                    response.add_ok(subscriber_ref, subscription_id)

                    # run callback method for subscriptions
                    if self._on_unsubscribe is not None:
                        bgt.add_task(self._on_unsubscribe, subscription)
                else:
                    response.add_error(subscription_id)
                
            except Exception:
                # respond with SubscriptionResponse Error for this subscription
                response.add_error(subscription_id)

        if self._datalog is not None:
            Datalog.create(self._datalog, response.xml(), {
                'method': req.method,
                'endpoint': str(req.url),
                'headers': dict(req.headers)
            }, self._service_participant_ref, 'IN', 'TerminateSubscriptionRequest', 'Response')

        return Response(content=response.xml(), media_type='application/xml')

    async def _request(self, req: Request, bgt: BackgroundTasks) -> Response:
        xml = await req.body()
            
        if self._datalog is not None:
            Datalog.create(self._datalog, xml, {
                'method': req.method,
                'endpoint': str(req.url),
                'headers': dict(req.headers)
            }, self._service_participant_ref, 'IN', 'SituationExchangeRequest', 'Request')
        
        request = xml2siri_request(xml)

        # run callback method for requests
        if self._on_request is not None:
            bgt.add_task(self._on_request)

        delivery = SituationExchangeDelivery(self._service_participant_ref, None)
        for _, situation in self._local_node_database.get_situations().items():
            delivery.add_situation(situation)

        if self._datalog is not None:
            Datalog.create(self._datalog, delivery.xml(), {
                'method': req.method,
                'endpoint': str(req.url),
                'headers': dict(req.headers)
            }, self._service_participant_ref, 'IN', 'SituationExchangeRequest', 'Response')

        return Response(content=delivery.xml(), media_type='application/xml')



import uuid

from abc import ABC
from lxml.etree import cleanup_namespaces
from lxml.etree import tostring
from lxml.objectify import deannotate
from lxml.objectify import fromstring
from lxml.objectify import Element

from .isotime import timestamp
from .model import Subscription


class SiriDelivery(ABC):

    def __init__(self):
        self.Siri = Element('Siri', xmlns='http://www.siri.org.uk/siri', version='2.1')

    def xml(self) -> str:
        deannotate(self.Siri)
        cleanup_namespaces(self.Siri)

        return tostring(self.Siri, pretty_print=True, xml_declaration=True, encoding='UTF-8')
    

class ServiceDelivery(SiriDelivery):

    def __init__(self, producer_ref: str, more_data=False):
        self.Siri = Element('Siri', xmlns='http://www.siri.org.uk/siri', version='2.1')

        self.Siri.ServiceDelivery = Element('ServiceDelivery')
        self.Siri.ServiceDelivery.ResponseTimestamp = timestamp()
        self.Siri.ServiceDelivery.ProducerRef = producer_ref
        self.Siri.ServiceDelivery.ResponseMessageIdentifier = str(uuid.uuid4())
        self.Siri.ServiceDelivery.Status = True
        self.Siri.ServiceDelivery.MoreData = more_data


class SituationExchangeDelivery(ServiceDelivery):

    def __init__(self, producer_ref: str, subscription: Subscription):
        super().__init__(producer_ref)

        self.Siri.ServiceDelivery.SituationExchangeDelivery = Element('SituationExchangeDelivery', version='2.1')
        self.Siri.ServiceDelivery.SituationExchangeDelivery.ResponseTimestamp = timestamp()

        if subscription is not None:
            self.Siri.ServiceDelivery.SituationExchangeDelivery.SubscriberRef = subscription.subscriber
            self.Siri.ServiceDelivery.SituationExchangeDelivery.SubscriptionRef = subscription.id
            
        self.Siri.ServiceDelivery.SituationExchangeDelivery.Situations = Element('Situations')

    def add_situation(self, situation):
        self.Siri.ServiceDelivery.SituationExchangeDelivery.Situations.append(situation)


def xml2siri_delivery(xml: str) -> SiriDelivery:
    request = SiriDelivery()
    request.Siri = fromstring(xml)

    return request


import logging
import sys
import time

from vdv736.subscriber import Subscriber
from vdv736.delivery import SiriDelivery
from vdv736.sirixml import exists as sirixml_exists
from vdv736.sirixml import get_elements as sirixml_get_elements

logging.basicConfig(level=logging.INFO, format="[%(asctime)s] %(levelname)s: %(message)s")

def on_delivery(delivery: SiriDelivery) -> None:
    print('Delivery callback called...')

    if sirixml_exists(delivery, 'Siri.ServiceDelivery.SituationExchangeDelivery.Situations'):
        pt_situations = sirixml_get_elements(delivery, 'Siri.ServiceDelivery.SituationExchangeDelivery.Situations.PtSituationElement')
        print(f"Found {len(pt_situations)} situations in this delivery")



with Subscriber('PY_TEST_SUBSCRIBER', './demo/demo_participants.yaml') as subscriber:

    if len(sys.argv) > 1 and sys.argv[1] == 'ps':

        subscriber.set_callbacks(on_delivery)

        time.sleep(15)
        sid = subscriber.subscribe('PY_TEST_PUBLISHER')
        time.sleep(10)
        subscriber.status(sid)
        time.sleep(10)
        subscriber.unsubscribe(sid)

        sid = subscriber.subscribe('PY_TEST_PUBLISHER')

        while True:
            time.sleep(30)
            subscriber.status()

    else:

        time.sleep(25)
        subscriber.request('PY_TEST_PUBLISHER')
        print(subscriber.get_situations())

        while True:
            pass


import datetime

def timestamp(additional_seconds=0) -> str:
    ts = datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)
    ts = ts + datetime.timedelta(seconds=additional_seconds)

    return ts.isoformat()

def interval(years: int, months: int, days: int, hours: int, minutes: int, seconds: int) -> str:
    result = 'P'

    if years > 0:
        result = result + f"{years}Y"
    
    if months > 0:
        result = result + f"{months}M"

    if days > 0:
        result = result + f"{days}D"
    
    result = result + 'T'

    if hours > 0:
        result = result + f"{hours}H"
    
    if minutes > 0:
        result = result + f"{minutes}M"

    if seconds > 0:
        result = result + f"{seconds}S"

    return result




import logging
import os
import platform
import sqlite3
import tempfile

from .model import PublicTransportSituation
from .model import Subscription


class LocalNodeDatabase:

    def __init__(self, name):
        # create logger
        self._logger = logging.getLogger('uvicorn')

        # connect to database
        tempdir = "/tmp" if platform.system() == "Darwin" else tempfile.gettempdir()
        self._filename = os.path.join(tempdir, name)

        self._connection = sqlite3.connect(self._filename, check_same_thread=False)
        self._connection.row_factory = sqlite3.Row

        # init required tables if not already done
        cursor = self._connection.cursor()
        cursor.execute("CREATE TABLE IF NOT EXISTS subscriptions (id TEXT NOT NULL PRIMARY KEY, serialized TEXT NOT NULL)")
        cursor.execute("CREATE TABLE IF NOT EXISTS situations (id TEXT NOT NULL PRIMARY KEY, serialized TEXT NOT NULL)")
        self._connection.commit()

    def get_subscriptions(self) -> dict[str, Subscription]:
        cursor = self._connection.cursor()
        cursor.execute("SELECT * FROM subscriptions")

        subscriptions = dict()
        for s in cursor.fetchall():
            subscription = Subscription.unserialize(s['serialized'])
            subscriptions[s['id']] = subscription

        return subscriptions

    def add_subscription(self, subscription_id: str, subscription: Subscription) -> bool:
        try:
            serialized = Subscription.serialize(subscription)
            
            cursor = self._connection.cursor()
            cursor.execute("INSERT INTO subscriptions (id, serialized) VALUES (?, ?)", (
                subscription_id,
                serialized,
            ))
            self._connection.commit()

            return True
        except sqlite3.Error as ex:
            self._logger.error(ex)
            return False
        
    def update_subscription(self, subscription_id: str, subscription: Subscription) -> bool:
        try:
            serialized = Subscription.serialize(subscription)
            
            cursor = self._connection.cursor()
            cursor.execute("UPDATE subscriptions SET serialized = ? WHERE id = ?", (
                serialized, 
                subscription_id
            ))
            self._connection.commit()

            return True
        except sqlite3.Error as ex:
            self._logger.error(ex)
            return False

    def remove_subscription(self, subscription_id: str) -> bool:
        try:
            cursor = self._connection.cursor()
            cursor.execute("DELETE FROM subscriptions WHERE id = ?", (subscription_id,))
            self._connection.commit()

            return True
        except sqlite3.Error as ex:
            self._logger.error(ex)
            return False

    def get_situations(self) -> dict[str, PublicTransportSituation]:
        cursor = self._connection.cursor()
        cursor.execute("SELECT * FROM situations")

        subscriptions = dict()
        for s in cursor.fetchall():
            situation = PublicTransportSituation.unserialize(s['serialized'])
            subscriptions[s['id']] = situation

        return subscriptions
    
    def add_situation(self, situation_id, situation: PublicTransportSituation) -> bool:
        try:
            serialized = PublicTransportSituation.serialize(situation)
            
            cursor = self._connection.cursor()
            cursor.execute("INSERT INTO situations (id, serialized) VALUES (?, ?)", (
                situation_id,
                serialized,
            ))
            self._connection.commit()

            return True
        except sqlite3.Error as ex:
            self._logger.error(ex)
            return False
        
    def update_situation(self, situation_id, situation: PublicTransportSituation) -> bool:
        try:
            serialized = PublicTransportSituation.serialize(situation)
            
            cursor = self._connection.cursor()
            cursor.execute("UPDATE situations SET serialized = ? WHERE id = ?", (
                serialized, 
                situation_id,
            ))
            self._connection.commit()

            return True
        except sqlite3.Error as ex:
            self._logger.error(ex)
            return False
        
    def add_or_update_situation(self, situation_id, situation: PublicTransportSituation) -> bool:
        try:
            cursor = self._connection.cursor()
            cursor.execute("SELECT COUNT(*) AS 'count' FROM situations WHERE id = ?", (situation_id,))
            count = cursor.fetchone()['count']

            if count == 0:
                self.add_situation(situation_id, situation)
            else:
                self.update_situation(situation_id, situation)

            return True
        except sqlite3.Error as ex:
            self._logger.error(ex)
            return False

    def remove_situation(self, situation_id) -> bool:
        try:
            cursor = self._connection.cursor()
            cursor.execute("DELETE FROM situations WHERE id = ?", (situation_id,))
            self._connection.commit()

            return True
        except sqlite3.Error as ex:
            self._logger.error(ex)
            return False
        
    def close(self, remove=False) -> None:
        self._connection.close()

        if remove == True:
            try:
                os.remove(self._filename)
            except PermissionError as ex:
                self._logger.error(ex)


def local_node_database(name: str) -> LocalNodeDatabase:
    if not name.endswith('.db3'):
        name = name + '.db3'

    return LocalNodeDatabase(name)


import os
import unittest
import unittest.mock

from fastapi.testclient import TestClient

from vdv736.publisher import PublisherEndpoint
from vdv736.response import xml2siri_response
from vdv736.sirixml import get_elements as sirixml_get_elements

class SubscriberEndpoint_Test(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.endpoint = PublisherEndpoint('TEST')

        cls.client = TestClient(cls.endpoint.create_endpoint(
            'TEST', 
            '/vdv736'
        ))

        return super().setUpClass()
    
    def test_SampleSubscriptionRequest(self):
        on_subscribe_callback = unittest.mock.Mock()
        on_unsubscribe_callback = unittest.mock.Mock()

        self.endpoint.set_callbacks(on_subscribe_callback=on_subscribe_callback, on_unsubscribe_callback=on_unsubscribe_callback)
        
        xml_filename = os.path.join(os.path.dirname(__file__), 'data/xml/SampleSubscriptionRequest.xml')
        with open(xml_filename, 'r') as xml_file:
            response = self.client.post('/vdv736', content=xml_file.read())

            self.assertEqual(200, response.status_code)
            on_subscribe_callback.assert_called()
            on_unsubscribe_callback.assert_not_called()

            siri_response = xml2siri_response(response.content)
            self.assertEqual(True, siri_response.Siri.SubscriptionResponse.ResponseStatus.Status)

    def test_SampleTerminateSpecificSubscriptionRequest(self):
        on_subscribe_callback = unittest.mock.Mock()
        on_unsubscribe_callback = unittest.mock.Mock()

        self.endpoint.set_callbacks(on_subscribe_callback=on_subscribe_callback, on_unsubscribe_callback=on_unsubscribe_callback)
        
        xml_filename = os.path.join(os.path.dirname(__file__), 'data/xml/SampleTerminateSpecificSubscriptionRequest.xml')
        with open(xml_filename, 'r') as xml_file:
            response = self.client.post('/vdv736', content=xml_file.read())

            self.assertEqual(200, response.status_code)
            on_subscribe_callback.assert_not_called()
            on_unsubscribe_callback.assert_called()

            siri_response = xml2siri_response(response.content)
            termination_response_status = sirixml_get_elements(siri_response, 'Siri.TerminationSubscriptionResponse.TerminationResponseStatus')
            for trs in termination_response_status:
                self.assertEqual(True, trs.Status)

    def test_SampleTerminateSubscriptionRequest(self):
        on_subscribe_callback = unittest.mock.Mock()
        on_unsubscribe_callback = unittest.mock.Mock()

        self.endpoint.set_callbacks(on_subscribe_callback=on_subscribe_callback, on_unsubscribe_callback=on_unsubscribe_callback)
        
        xml_filename = os.path.join(os.path.dirname(__file__), 'data/xml/SampleTerminateSubscriptionRequest.xml')
        with open(xml_filename, 'r') as xml_file:
            response = self.client.post('/vdv736', content=xml_file.read())

            self.assertEqual(200, response.status_code)
            on_subscribe_callback.assert_not_called()
            # on_unsubscribe_callback.assert_called() # cannot be called here, since test_SampleTerminateSpecificSubscriptionRequest already terminated every subscription on the test publisher

            siri_response = xml2siri_response(response.content)
            termination_response_status = sirixml_get_elements(siri_response, 'Siri.TerminationSubscriptionResponse.TerminationResponseStatus')
            
            self.assertEqual(0, len(termination_response_status))        

    def test_InvalidXmlData(self):
        xml_filename = os.path.join(os.path.dirname(__file__), 'data/xml/InvalidXmlData.xml')
        with open(xml_filename, 'r') as xml_file:
            response = self.client.post('/vdv736', content=xml_file.read())

            self.assertEqual(400, response.status_code)

    def test_InvalidEndpoint(self):
        response = self.client.post('/someInvalidEndpoint')
        self.assertEqual(404, response.status_code)

    def test_InvalidMethod(self):
        response = self.client.get('/vdv736')
        self.assertEqual(405, response.status_code)

    @classmethod
    def tearDownClass(cls):

        # close local node database in order to cleanup endpoint data
        # note: normally, this would be done by the Subscriber class, as the endpoint
        # is started as thread of the subscriber!
        cls.endpoint._local_node_database.close(True)

        return super().tearDownClass()
            


import re

def exists(obj, path):
    path = path.split('.')
    
    level0 = path[0]
    level1 = '.'.join(path[1:])
    
    return obj is not None and hasattr(obj, level0) if len(path) == 1 else exists(getattr(obj, level0), level1) if hasattr(obj, level0) else False

def get_elements(obj, path):
    if exists(obj, path):
        path = path.split('.')
        
        destination = obj
        for element in path:
            destination = getattr(destination, element)
            
        return destination

    return list()
    
def get_value(obj, path, default=None):
    if exists(obj, path):
        path = path.split('.')
        
        destination = obj
        for element in path:
            destination = getattr(destination, element)
            
        if hasattr(destination, 'text'):
            return destination.text
        else:
            return default
        
    return default

def get_attribute(obj, path, default=None):
    regex = re.compile('\.(?![^{]*})')
    
    objectpath = regex.split(path)
    objectpath = '.'.join(objectpath[:-1])
    
    if exists(obj, objectpath):
        path = regex.split(path)
        
        destination = obj
        for element in path[:-1]:
            destination = getattr(destination, element)
            
        if hasattr(destination, 'attrib') and path[-1] in destination.attrib:
            return destination.attrib[(path[-1])]
        else:
            return default
            
    return default


import unittest

from lxml.objectify import fromstring

from vdv736.sirixml import get_attribute as sirixml_get_attribue

class Sirixml_Test(unittest.TestCase):

    def test_get_attribute_withoutNamespaces(self):

        xml = """
        <Siri xmlns="http://www.siri.org.uk/siri" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" version="2.0">
        </Siri>
        """

        obj = lambda: None
        obj.Siri = fromstring(xml)

        self.assertEqual('2.0', sirixml_get_attribue(obj, 'Siri.version'))
        self.assertEqual(None, sirixml_get_attribue(obj, 'Siri.invalid_attribute'))

    def test_get_attribute_withNamespace(self):

        xml = """
        <Siri xmlns="http://www.siri.org.uk/siri" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" version="2.0">
        <Situation>
        <Summary xml:lang="DE">This is a sample element</Summary>
        </Situation>
        </Siri>
        """

        obj = lambda: None
        obj.Siri = fromstring(xml)

        self.assertEqual('DE', sirixml_get_attribue(obj, 'Siri.Situation.Summary.{http://www.w3.org/XML/1998/namespace}lang'))
        self.assertEqual(None, sirixml_get_attribue(obj, 'Siri.Situation.Summary.lang'))




from abc import ABC
from lxml.etree import cleanup_namespaces
from lxml.etree import tostring
from lxml.objectify import deannotate
from lxml.objectify import fromstring
from lxml.objectify import Element

from .isotime import timestamp
from .isotime import interval


class SiriResponse(ABC):

    def __init__(self):
        self.Siri = Element('Siri', xmlns='http://www.siri.org.uk/siri', version='2.0')

    def xml(self) -> str:
        deannotate(self.Siri)
        cleanup_namespaces(self.Siri)

        return tostring(self.Siri, pretty_print=True, xml_declaration=True, encoding='UTF-8') 
    

class CheckStatusResponse(SiriResponse):

     def __init__(self, service_started_time: str):
         super().__init__()

         self.Siri.CheckStatusResponse = Element('CheckStatusResponse')
         self.Siri.CheckStatusResponse.ResponseTimestamp = timestamp()
         self.Siri.CheckStatusResponse.Status = True
         self.Siri.CheckStatusResponse.ValidUntil = ''
         self.Siri.CheckStatusResponse.ShortestPossibleCycle = interval(0, 0, 0, 0, 1, 0)
         self.Siri.CheckStatusResponse.ServiceStartedTime = service_started_time


class SubscriptionResponse(SiriResponse):

    def __init__(self, responder_ref: str, service_started_time: str = None):
        super().__init__()

        self.Siri.SubscriptionResponse = Element('SubscriptionResponse')
        self.Siri.SubscriptionResponse.ResponseTimestamp = Element('ResponseTimestamp')
        self.Siri.SubscriptionResponse.ResponderRef = Element('ResponderRef')
        self.Siri.SubscriptionResponse.ResponseStatus = Element('ResponseStatus')

        self.Siri.SubscriptionResponse.ResponseTimestamp = timestamp()
        self.Siri.SubscriptionResponse.ResponderRef = responder_ref

        if service_started_time is not None:
            self.Siri.SubscriptionResponse.ResponseStatus.ServiceStartedTime = service_started_time        

    def ok(self, subscription_id: str, subscription_termination: str):

        self.Siri.SubscriptionResponse.ResponseStatus.ResponseTimestamp = timestamp()
        self.Siri.SubscriptionResponse.ResponseStatus.SubscriptionRef = subscription_id
        self.Siri.SubscriptionResponse.ResponseStatus.Status = True
        self.Siri.SubscriptionResponse.ResponseStatus.ValidUntil = subscription_termination
        self.Siri.SubscriptionResponse.ResponseStatus.ShortestPossibleCycle = interval(0, 0, 0, 0, 1, 0)

    def error(self, subscription_id: str):

        self.Siri.SubscriptionResponse.ResponseStatus.ResponseTimestamp = timestamp()
        self.Siri.SubscriptionResponse.ResponseStatus.SubscriptionRef = subscription_id
        self.Siri.SubscriptionResponse.ResponseStatus.Status = False

        self.Siri.SubscriptionResponse.ResponseStatus.ErrorCondition = Element('ErrorCondition')


class TerminateSubscriptionResponse(SiriResponse):

    def __init__(self, responder_ref: str):
        super().__init__()

        self.Siri.TerminationSubscriptionResponse = Element('TerminationSubscriptionResponse')
        self.Siri.TerminationSubscriptionResponse.ResponseTimestamp = timestamp()
        self.Siri.TerminationSubscriptionResponse.ResponderRef = responder_ref

    def ok(self, subscriber_ref: str, subscription_id: str):
        self.add_ok(subscriber_ref, subscription_id)

    def error(self, subscription_id: str):
        self.add_error(subscription_id)

    def add_ok(self, subscriber_ref: str, subscription_id: str):
        
        termination_response_status = Element('TerminationResponseStatus')
        termination_response_status.ResponseTimestamp = timestamp()
        termination_response_status.SubscriberRef = subscriber_ref
        termination_response_status.SubscriptionRef = subscription_id
        termination_response_status.Status = True

        self.Siri.TerminationSubscriptionResponse.append(termination_response_status)

    def add_error(self, subscription_id: str):

        termination_response_status = Element('TerminationResponseStatus')
        termination_response_status.SubscriptionRef = subscription_id
        termination_response_status.Status = False

        termination_response_status.ErrorCondition = Element('ErrorCondition')
        termination_response_status.ErrorCondition.OtherError = Element('OtherError')

        self.Siri.TerminationSubscriptionResponse.append(termination_response_status)


class DataReceivedAcknowledgement(SiriResponse):

    def __init__(self, consumer_ref, request_message_ref: str):
        super().__init__()

        self.Siri.DataReceivedAcknowledgement = Element('DataReceivedAcknowledgement')
        self.Siri.DataReceivedAcknowledgement.ResponseTimestamp = timestamp()
        self.Siri.DataReceivedAcknowledgement.ConsumerRef = consumer_ref
        self.Siri.DataReceivedAcknowledgement.RequestMessageRef = request_message_ref

    def ok(self):
        self.Siri.DataReceivedAcknowledgement.Status = True

    def error(self):
        self.Siri.DataReceivedAcknowledgement.Status = False

def xml2siri_response(xml: str) -> SiriResponse:
    response = SiriResponse()
    response.Siri = fromstring(xml)

    return response


import os
import unittest
import unittest.mock

from fastapi.testclient import TestClient

from vdv736.model import Subscription
from vdv736.subscriber import SubscriberEndpoint
from vdv736.response import xml2siri_response

class SubscriberEndpoint_Test(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.endpoint = SubscriberEndpoint('TEST')

        # setup a virtual subscription, otherwise the subscriber will fail
        # see #21 for details
        virtual_subscription = Subscription()
        virtual_subscription.remote_service_participant_ref = 'TEST-PUBLISHER'

        cls.endpoint._local_node_database.add_subscription('1', virtual_subscription)

        # create test client
        cls.client = TestClient(cls.endpoint.create_endpoint(
            '/vdv736'
        ))

        return super().setUpClass()
    
    def test_SampleServiceDelivery(self):
        xml_filename = os.path.join(os.path.dirname(__file__), 'data/xml/SampleServiceDelivery.xml')
        with open(xml_filename, 'r') as xml_file:
            response = self.client.post('/vdv736', content=xml_file.read())

            self.assertEqual(200, response.status_code)

            siri_response = xml2siri_response(response.content)
            self.assertEqual(True, siri_response.Siri.DataReceivedAcknowledgement.Status)

    def test_SampleServiceDeliveryWithCallbacks(self):
        on_delivery_callback = unittest.mock.Mock()
            
        self.endpoint.set_callbacks(on_delivery_callback)
        
        xml_filename = os.path.join(os.path.dirname(__file__), 'data/xml/SampleServiceDelivery.xml')
        with open(xml_filename, 'r') as xml_file:
            
            response = self.client.post('/vdv736', content=xml_file.read())

            self.assertEqual(200, response.status_code)
            on_delivery_callback.assert_called()

            siri_response = xml2siri_response(response.content)
            self.assertEqual(True, siri_response.Siri.DataReceivedAcknowledgement.Status)

        self.endpoint.set_callbacks(None)

    def test_SampleServiceDeliveryWithInvalidPublisher(self):
        xml_filename = os.path.join(os.path.dirname(__file__), 'data/xml/SampleServiceDeliveryWithInvalidPublisher.xml')
        with open(xml_filename, 'r') as xml_file:
            response = self.client.post('/vdv736', content=xml_file.read())

            self.assertEqual(401, response.status_code)
    
    def test_InvalidXmlData(self):
        xml_filename = os.path.join(os.path.dirname(__file__), 'data/xml/InvalidXmlData.xml')
        with open(xml_filename, 'r') as xml_file:
            response = self.client.post('/vdv736', content=xml_file.read())

            self.assertEqual(400, response.status_code)

    def test_InvalidEndpoint(self):
        response = self.client.post('/someInvalidEndpoint')
        self.assertEqual(404, response.status_code)

    def test_InvalidMethod(self):
        response = self.client.get('/vdv736')
        self.assertEqual(405, response.status_code)

    @classmethod
    def tearDownClass(cls):

        # close local node database in order to cleanup endpoint data
        # note: normally, this would be done by the Subscriber class, as the endpoint
        # is started as thread of the subscriber!
        cls.endpoint._local_node_database.close(True)

        return super().tearDownClass()
            


import yaml

class ParticipantConfig:

    def __init__(self, participant_config_filename):

        self.participants = dict()

        with open(participant_config_filename) as participant_config_file:
                default_config = {
                    'host': '127.0.0.1',
                    'port': 9090,
                    'protocol': 'http',
                    'single_endpoint': None,
                    'status_endpoint': '/status',
                    'subscribe_endpoint': '/subscribe',
                    'unsubscribe_endpoint': '/unsubscribe',
                    'request_endpoint': '/request',
                    'delivery_endpoint': '/delivery'
                }

                configs = yaml.safe_load(participant_config_file)
                for participant_id, participant_config in configs.items():
                    merged_config = self._merge_config(
                        default_config, 
                        participant_config
                    )

                    unknown_keys = merged_config.keys() - default_config.keys()
                    if len(unknown_keys) > 0:
                         raise ValueError(f"unknown participant config key(s) {unknown_keys}")

                    self.participants[participant_id] = merged_config

    def _merge_config(self, defaults, actual):
        if isinstance(defaults, dict) and isinstance(actual, dict):
            return {k: self._merge_config(defaults.get(k, {}), actual.get(k, {})) for k in set(defaults) | set(actual)}
        
        return actual if actual or actual == None else defaults


import os
import responses
import unittest
import unittest.mock

from vdv736.subscriber import Subscriber
from vdv736.request import InvalidMethodError

class SubscriberDirectRequest_Test(unittest.TestCase):

    def setUp(self):
        xml_filename = os.path.join(os.path.dirname(__file__), 'data/xml/SampleServiceDelivery.xml')
        with open(xml_filename, 'r') as xml_file:
            xml_content = xml_file.read()
        
        self.responder = responses.RequestsMock(assert_all_requests_are_fired=False)
        self.responder.start()

        self.responder.add(
            responses.POST,
            'http://127.0.0.1:9091/request',
            body=xml_content,
            content_type='application/xml',
            status=200
        )

        self.responder.add(
            responses.GET,
            'http://127.0.0.1:9091/request',
            body=xml_content,
            content_type='application/xml',
            status=200
        )

    def test_DirectRequestResponse(self):

        configfile = os.path.join(os.path.dirname(__file__), 'data/yaml/participants_testconfig.yaml')        
        with Subscriber('PY_TEST_SUBSCRIBER', configfile, publish_subscribe=False) as subscriber:
            self.assertEqual(True, subscriber.request('PY_TEST_PUBLISHER'))
            self.assertEqual(1, len(subscriber.get_situations()))

    def test_DirectRequestResponseWithCallback(self):
  
        on_delivery_callback = unittest.mock.Mock()
        
        configfile = os.path.join(os.path.dirname(__file__), 'data/yaml/participants_testconfig.yaml')
        with Subscriber('PY_TEST_SUBSCRIBER', configfile, publish_subscribe=False) as subscriber:
            subscriber.set_callbacks(on_delivery_callback)

            self.assertEqual(True, subscriber.request('PY_TEST_PUBLISHER'))
            on_delivery_callback.assert_called()

    def test_DirectRequestResponseUsingGET(self):

        configfile = os.path.join(os.path.dirname(__file__), 'data/yaml/participants_testconfig.yaml')        
        with Subscriber('PY_TEST_SUBSCRIBER', configfile, publish_subscribe=False) as subscriber:
            self.assertEqual(True, subscriber.request('PY_TEST_PUBLISHER', method='GET'))
            self.assertEqual(1, len(subscriber.get_situations()))

    def test_DirectRequestResponseUsingInvalidMethod(self):

        configfile = os.path.join(os.path.dirname(__file__), 'data/yaml/participants_testconfig.yaml')        
        with Subscriber('PY_TEST_SUBSCRIBER', configfile, publish_subscribe=False) as subscriber:
            with self.assertRaises(InvalidMethodError):
                subscriber.request('PY_TEST_PUBLISHER', method='TESTMETHOD')
            
    def tearDown(self):
        self.responder.stop()
        self.responder.reset()

        return super().tearDownClass()


from abc import ABC
from lxml.etree import cleanup_namespaces
from lxml.etree import tostring
from lxml.objectify import deannotate
from lxml.objectify import fromstring
from lxml.objectify import Element

from .isotime import timestamp
from .isotime import interval
from .model import Subscription


class InvalidMethodError(Exception):
    pass

class SiriRequest(ABC):

    def __init__(self):
        self.Siri = Element('Siri', xmlns='http://www.siri.org.uk/siri', version='2.0')
    
    def xml(self) -> str:
        deannotate(self.Siri)
        cleanup_namespaces(self.Siri)

        return tostring(self.Siri, pretty_print=True, xml_declaration=True, encoding='UTF-8')


class CheckStatusRequest(SiriRequest):

    def __init__(self, subscription: Subscription):
        super().__init__()

        self.Siri.CheckStatusRequest = Element('CheckStatusRequest', version='2.0')
        self.Siri.CheckStatusRequest.RequestTimestamp = timestamp()
        self.Siri.CheckStatusRequest.RequestorRef = subscription.subscriber


class SubscriptionRequest(SiriRequest):

    def __init__(self, subscriber_ref: str):
        super().__init__()

        self.Siri.SubscriptionRequest = Element('SubscriptionRequest')
        self.Siri.SubscriptionRequest.RequestTimestamp = timestamp()
        self.Siri.SubscriptionRequest.RequestorRef = subscriber_ref


class ServiceRequest(SiriRequest):

    def __init__(self, subscriber_ref: str):
        super().__init__()

        self.Siri.ServiceRequest = Element('ServiceRequest')
        self.Siri.ServiceRequest.RequestTimestamp = timestamp()
        self.Siri.ServiceRequest.RequestorRef = subscriber_ref


class TerminateSubscriptionRequest(SiriRequest):

    def __init__(self, subscriber_ref: str):
        super().__init__()

        self.Siri.TerminateSubscriptionRequest = Element('TerminateSubscriptionRequest')
        self.Siri.TerminateSubscriptionRequest.RequestTimestamp = timestamp()
        self.Siri.TerminateSubscriptionRequest.RequestorRef = subscriber_ref

    def subscription(self, subscriber_ref: str, subscription_id: str) -> None:
        self.Siri.TerminateSubscriptionRequest.SubscriberRef = subscriber_ref
        self.Siri.TerminateSubscriptionRequest.SubscriptionRef = subscription_id

    def all(self):
        self.Siri.TerminateSubscriptionRequest.All = Element('All')


class SituationExchangeSubscriptionRequest(SubscriptionRequest):

    def __init__(self, subscription: Subscription):
        super().__init__(subscription.subscriber)

        self.Siri.SubscriptionRequest.SituationExchangeSubscriptionRequest = Element('SituationExchangeSubscriptionRequest')
        self.Siri.SubscriptionRequest.SituationExchangeSubscriptionRequest.SubscriberRef = subscription.subscriber
        self.Siri.SubscriptionRequest.SituationExchangeSubscriptionRequest.SubscriptionIdentifier = subscription.id
        self.Siri.SubscriptionRequest.SituationExchangeSubscriptionRequest.InitialTerminationTime = subscription.termination

        self.Siri.SubscriptionRequest.SituationExchangeSubscriptionRequest.SituationExchangeRequest = Element('SituationExchangeRequest')
        self.Siri.SubscriptionRequest.SituationExchangeSubscriptionRequest.SituationExchangeRequest.RequestTimestamp = timestamp()


class SituationExchangeRequest(ServiceRequest):

    def __init__(self, subscriber_ref: str):
        super().__init__(subscriber_ref)

        self.Siri.ServiceRequest.SituationExchangeRequest = Element('SituationExchangeRequest', version='2.0')
        self.Siri.ServiceRequest.SituationExchangeRequest.RequestTimestamp = timestamp()


def xml2siri_request(xml: str) -> SiriRequest:
    request = SiriRequest()
    request.Siri = fromstring(xml)

    return request

        


import logging

from vdv736.sirixml import get_value as sirixml_get_value

from abc import ABC, abstractmethod
from datetime import datetime
from datetime import timezone


class BaseSituationHandler(ABC):

    @abstractmethod
    def handle_situation(cls, pts, chaining_result=True) -> bool:
        """Returns whether the situation object can be discarded or not."""
        pass

class SituationProgressHandler(BaseSituationHandler):

    @classmethod
    def handle_situation(cls, pts, chaining_result=True) -> bool:
        situation_id = sirixml_get_value(pts, 'SituationNumber')
        situation_progress = sirixml_get_value(pts, 'Progress', None)

        if situation_progress is None:
            logging.warning(f"Situation {situation_id} has not set Progress attribute!")
        
        if situation_progress == 'closed':
            return False
        elif situation_progress == 'closing':
            situation_versioned_at = sirixml_get_value(pts, 'VersionedAtTime', None)
            if situation_versioned_at is not None:
                situation_versioned_at = datetime.fromisoformat(situation_versioned_at.replace('Z', '+00:00'))
                situation_deprecation_difference_minutes = int((datetime.now(timezone.utc) - situation_versioned_at).total_seconds() // 60)

                if situation_deprecation_difference_minutes >= 5:
                    return False
                else:
                    return chaining_result
            else:
                return chaining_result
        else:
            return chaining_result


import unittest
import datetime

from vdv736.isotime import timestamp
from vdv736.isotime import interval


class Timestamp_Test(unittest.TestCase):
    def test_timestamp(self):
        seconds = [15, 60, 60 * 60, 60 * 60 * 24]

        for s in seconds:
            reference_timestamp = datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)
            reference_timestamp = reference_timestamp + datetime.timedelta(seconds=s)

            self.assertEqual(timestamp(s), reference_timestamp.isoformat())
            self.assertTrue(timestamp(s).endswith('+00:00') or timestamp(s).endswith('Z'), timestamp(s))

    def test_interval(self):
        self.assertEqual(interval(0, 0, 0, 1, 0, 0), 'PT1H')
        self.assertEqual(interval(1, 0, 5, 0, 10, 0), 'P1Y5DT10M')
        self.assertEqual(interval(0, 0, 0, 0, 0, 15), 'PT15S')
        self.assertEqual(interval(0, 0, 0, 0, 5, 0), 'PT5M')



import os
import responses
import unittest
import unittest.mock

from vdv736.delivery import xml2siri_delivery
from vdv736.delivery import SiriDelivery
from vdv736.handler import SituationProgressHandler
from vdv736.isotime import timestamp
from vdv736.sirixml import get_elements as sirixml_get_elements

class SubscriberDirectRequest_Test(unittest.TestCase):

    def _extract_first_situation(self, filename: str) -> object:
        xml_filename = os.path.join(os.path.dirname(__file__), filename)
        with open(xml_filename, 'rb') as xml_file:
            xml_content = xml_file.read()

        siri_delivery = xml2siri_delivery(xml_content)

        return sirixml_get_elements(siri_delivery, 'Siri.ServiceDelivery.SituationExchangeDelivery.Situations.PtSituationElement')[0]

    def test_DecideWhetherToDelete_WithClosingSituation(self):

        pts = self._extract_first_situation('data/xml/SampleServiceDeliveryWithClosingSituation.xml')

        handler = SituationProgressHandler()

        result = handler.handle_situation(pts)
        self.assertEqual(False, result)

    def test_DecideWhetherToDelete_WithRecentlyUpdatedClosingSituation(self):

        pts = self._extract_first_situation('data/xml/SampleServiceDeliveryWithClosingSituation.xml')
        pts.VersionedAtTime = timestamp(-120)

        handler = SituationProgressHandler()

        result = handler.handle_situation(pts)
        self.assertEqual(True, result)
        

        

    	


import json
import os

from datetime import datetime
from lxml.etree import fromstring
from lxml.etree import tostring

class Datalog:

    @classmethod
    def cleanup(cls, directory: str, ttl_hours: int = 24) -> None:
        if not os.path.exists(directory) or not os.path.isdir(directory):
            os.makedirs(directory)

        # look for old datalog files and remove them
        # for speed up, check for the filename not beginning with today instead of ressource consuming difference calculation
        today = datetime.now().strftime('%Y-%m-%d')
        for datalog_file in os.listdir(directory):

            # proceed only if the datalogfile is not from today
            if not datalog_file.startswith(today):
                datalog_timestamp = datalog_file.split('_')[0]
                datalog_timestamp = datetime.strptime(datalog_timestamp, '%Y-%m-%d-%H.%M.%S-%f')

                difference = (datetime.now() - datalog_timestamp).total_seconds()
                if difference > 60 * 60 * ttl_hours:
                    datalog_file = os.path.join(directory, datalog_file)
                    os.remove(datalog_file)
    @classmethod
    def create(cls, directory: str, data: str, meta: dict, *args) -> None:
        cls.cleanup(directory)
        
        if not os.path.exists(directory) or not os.path.isdir(directory):
            os.makedirs(directory)

        # generate new datalog file
        datalog_timestamp = datetime.now().strftime('%Y-%m-%d-%H.%M.%S-%f')
        datalog_filename = f"{datalog_timestamp}_{'-'.join([str(a) for a in args])}.xml"

        with open(os.path.join(directory, datalog_filename), 'wb') as datalog_file:
                try:
                    xml = tostring(fromstring(data), pretty_print=True)
                    datalog_file.write(xml)
                    datalog_file.write(cls._comment(meta))
                except Exception:
                    datalog_file.write(data)
                    datalog_file.write(cls._comment(meta))
                finally:
                    datalog_file.close()

    @classmethod
    def _comment(cls, metadata: dict) -> str:
        comment = "<!--\n"
        comment += "This message is created by datalog module - See https://gist.github.com/sebastianknopf/677d9ae9f213389e1c3f9aeb0b5e4055\n"
        comment += "\n"
        comment += "Request Meta Data:\n"
        comment += json.dumps(metadata, indent=4) + "\n"
        comment += "-->"

        return comment.encode('utf-8')


from typing import Any
from lxml.etree import tostring
from lxml.objectify import fromstring
from lxml.objectify import Element
from lxml.objectify import ObjectifiedElement


class Subscription:

    @classmethod
    def create(cls, id: str, host: str, port: int, protocol: str, subscriber: str, termination: str):

        obj = cls()
        obj.id = id
        obj.host = host
        obj.port = port
        obj.protocol = protocol
        obj.subscriber = subscriber
        obj.termination = termination

        return obj

    @classmethod
    def serialize(cls, obj) -> str:
        element = Element('subscription')
        element.id = obj.id
        element.host = obj.host
        element.port = obj.port
        element.protocol = obj.protocol
        element.subscriber = obj.subscriber
        element.termination = obj.termination

        element.remote_service_participant_ref = obj.remote_service_participant_ref
        element.remote_service_startup_time = obj.remote_service_startup_time

        element.single_endpoint = obj.single_endpoint
        element.status_endpoint = obj.status_endpoint
        element.subscribe_endpoint = obj.subscribe_endpoint
        element.unsubscribe_endpoint = obj.unsubscribe_endpoint

        return tostring(element)
    
    @classmethod
    def unserialize(cls, xml: str):
        element = fromstring(xml)

        obj = cls()
        obj.id = element.id.text
        obj.host = element.host.text
        obj.port = element.port.text
        obj.protocol = element.protocol.text
        obj.subscriber = element.subscriber.text
        obj.termination = element.termination.text

        obj.remote_service_participant_ref = element.remote_service_participant_ref.text
        obj.remote_service_startup_time = element.remote_service_startup_time.text

        obj.single_endpoint = element.single_endpoint.text
        obj.status_endpoint = element.status_endpoint.text
        obj.subscribe_endpoint = element.subscribe_endpoint.text
        obj.unsubscribe_endpoint = element.unsubscribe_endpoint.text

        return obj
    
    def __init__(self):
        self.id = None
        self.host = None
        self.port = None
        self.protocol = None
        self.subscriber = None
        self.termination = None

        self.remote_service_participant_ref = None
        self.remote_service_startup_time = None

        self.single_endpoint = None
        self.status_endpoint = '/status'
        self.subscribe_endpoint = '/subscribe'
        self.unsubscribe_endpoint = '/unsubscribe'


class PublicTransportSituation(ObjectifiedElement):
    
    @classmethod
    def create(cls, id: str):

        obj = cls()
        obj.tag = 'PtSituationElement'

        obj.SituationNumber = id

        return obj
    
    @classmethod
    def serialize(cls, obj) -> str:
        return tostring(obj)
    
    @classmethod
    def unserialize(cls, xml: str):
        obj = fromstring(xml)
        return obj
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __eq__(self, other):
        return self.equals(other)

    def __copy__(self):
        return PublicTransportSituation(super().__copy__())

    def __deepcopy__(self, memo=None):
        return PublicTransportSituation(super().__deepcopy__(memo))